@@ -19,7 +19,7 @@ use strict;			# Earlier it catches ISA and EXPORT.
 # this flag tells cpp to only output macros
 $C::Scan::MACROS_ONLY = '-dM';
 
-$C::Scan::VERSION = '0.70';
+$C::Scan::VERSION = '0.72';
 
 my (%keywords,%style_keywords);
 for (qw(asm auto break case char continue default do double else enum
@@ -92,11 +92,14 @@ my $recipes
       typedef_chunks => { filter => [ \&typedef_chunks, 'full_toplevel'], },
       typedefs_maybe => { filter => [ sub {[keys %{+shift}]}, 'typedef_hash'], },
       typedefs_whited => { filter => [ \&typedefs_whited,
-				      'full_sanitized', 'typedef_chunks'], },
+				      'full_sanitized', 'typedef_chunks',
+				      'keywords_rex'], },
       typedef_texts => { filter => [ \&typedef_texts,
 				      'full_text', 'typedef_chunks'], },
       typedef_hash => { filter => [ \&typedef_hash,
 				    'typedef_texts', 'typedefs_whited'], },
+      typedef_structs => { filter => [ \&typedef_structs,
+				       'typedef_hash'], },
       defines_maybe => { filter => [ \&defines_maybe, 'filename'], },
       defines_no_args => { prerequisites => ['defines_maybe'],
 			   output => sub { shift->{defines_maybe}->[0] }, },
@@ -123,6 +126,11 @@ my $recipes
       vdecls => { filter => [ \&from_chunks, 'vdecl_chunks', 'text'], },
       parsed_fdecls => { filter => [ \&do_declarations, 'fdecls', 
 				     'typedef_hash', 'keywords'], },
+      keywords_rex => { filter => [ sub { my @k = keys %{ shift() };
+					  local $" = '|';
+					  my $r = "(?:@k)";
+					  eval 'qr/$r/' or $r	# Older Perls
+					}, 'keywords'], },
     };
 
 sub from_chunks {
@@ -252,12 +260,15 @@ sub defines_full {
 
 sub typedef_texts {
   my ($txt, $chunks) = (shift, shift);
-  my ($b, $e, @out);
+  my ($b, $e, $in, @out);
   my @in = @$chunks;
   while (($b, $e) = splice @in, 0, 2) {
-    push @out, substr($txt, $b, $e - $b);
+    $in = substr($txt, $b, $e - $b);
+    # remove any remaining directives
+    $in =~ s/^ ( \s* \# .* ( \\ $ \n .* )* ) / ' ' x length($1)/xgem;
+    push @out, $in;
   }
-  [@out];
+  \@out;
 }
 
 sub typedef_hash_old {
@@ -355,6 +366,60 @@ sub typedef_chunks {		# Input is toplevel, output: starts and ends
   \@out;
 }
 
+sub typedef_structs {
+  my $typehash = shift;
+  my %structs;
+  while (my($key, $text) = each %$typehash) {
+    my $name = parse_struct($text->[0], \%structs);
+    $structs{$key} = defined($name) ? $structs{$name} : undef;
+  }
+  \%structs;
+}
+
+sub parse_struct {
+  my($in, $structs) = @_;
+  my($struct, @type, $word, $id, $structname);
+  ($structname, $in) = $in =~ /^ \s* ( (?: struct | union ) (?: \s+ \w+ )? ) \s* { \s* (.*) } \s* $/gisx or return;
+  $structname .= " _ANON" unless $structname =~ /\s/;
+  $structname .= " 0" if exists $structs->{$structname};
+  $structname =~ s/(\d+$)/$1 + 1/e while exists $structs->{$structname};
+  while ($in =~ /\G\s*(\S+?)(\s+|\b\s*)/g) {
+    $word = $1;
+    if ($word eq ';') {
+      next unless @type;
+      $id = pop @type;
+      @type = (join ' ', @type ? @type : 'int'); # or is this an error?
+      push @$struct, [ $type[0], $id ];
+      pop @type;
+    } elsif ($word eq ',') {
+      $id = pop @type;
+      @type = (join ' ', @type ? @type : 'int'); # or is this an error?
+      push @$struct, [ $type[0], $id ];
+    } elsif ($word eq '{') {
+      # we have an embedded struct or union, I hope
+      my $b = pos $in;
+      matchingbrace($in);
+      my $e = pos $in;
+      my $chunk = join ' ', @type, $word, substr($in, $b, pos($in) - $b);
+      if (@type && ($type[0] eq 'struct' || $type[0] eq 'union') && @type <= 2) {
+	@type = parse_struct($chunk, $structs);
+	pos($in) = $e;
+      } else {
+	warn "panic: not a struct or union; ignoring '$chunk'.\n";
+      }
+    } else {
+      push @type, $word;
+    }
+  }
+  if (@type) {
+    $id = pop @type;
+    @type = (join ' ', @type ? @type : 'int'); # or is this an error?
+    push @$struct, [ $type[0], $id ];
+  }
+  $structs->{$structname} = $struct;
+  $structname;
+}
+
 # The output is the list of list of inline chunks and list of
 # declaration chunks.
 
@@ -391,11 +456,13 @@ sub functions_in {		# The arg is text without type declarations.
 	push @mdecls, $b + $b1, $e;
       } else  {			# Non-multiple.
 	my $isvar = 1;
+	# Since leading \s* is not optimized, this is quadratic!
 	$chunk =~ s{
-		    \s* ( ( const
-			    | __attribute__ \s* \( \s* \)
-			  ) \s* )* ( ; \s* )? \Z # Strip from the end
+		     ( ( const
+			 | __attribute__ \s* \( \s* \)
+		       ) \s* )* ( ; \s* )? \Z # Strip from the end
 		   }()x;
+	$chunk =~ s/\s*\Z//;
 	if ($chunk =~ /\)\Z/) { # Function declaration ends on ")"!
 	  if ($chunk !~ m{ 
 			  \( .* \( # Multiple parenths
@@ -423,7 +490,7 @@ sub typedefs_whited {		# Input is sanitized text, and list of beg/end.
   my ($b, $e);
   while ($b = shift @lst) {
     $e = shift @lst;
-    push @out, whited_decl(substr $_[0], $b, $e - $b);
+    push @out, whited_decl($_[2], substr $_[0], $b, $e - $b);
   }
   \@out;
 }
@@ -437,6 +504,7 @@ sub typedefs_whited {		# Input is sanitized text, and list of beg/end.
 # Now out of several words in a row the last one is a newly defined type.
 
 sub whited_decl {		# Input is sanitized.
+  my $keywords_rex = shift;
   my $in = shift;		# Text of a declaration
   my $rest  = $in;
   my $out  = $in;		# Whited out $in
@@ -468,11 +536,13 @@ sub whited_decl {		# Input is sanitized.
       pos $out = $att_pos_end;
   }
 
-  # Remove arguments of functions (heuristics only):
+  # Remove arguments of functions (heuristics only).
+  # These things (start) arglist of a declared function:
   # paren word comma
   # paren word space non-paren
+  # paren keyword paren
   # start a list of arguments. (May be "cdecl *myfunc"?) XXXXX ?????
-  while ( $out =~ /(\(\s*\w+(,|\s+[^\)\s]))/g ) {
+  while ( $out =~ /(\(\s*(\w+(,|\s+[^\)\s])|$keywords_rex\s*\)))/g ) {
     my $pos_start = pos($out) - length($1);
     pos $out = $pos_start + 1;
     matchingbrace($out);
@@ -480,7 +550,7 @@ sub whited_decl {		# Input is sanitized.
       = ' ' x (pos($out) - 2 - $pos_start);
   }
   # Remove array specifiers
-  $out =~ s/(\[[\w\s]*\])/ ' ' x length $1 /ge;
+  $out =~ s/(\[[\w\s\+]*\])/ ' ' x length $1 /ge;
   my $tout = $out;
   # Several words in a row cannot be new typedefs, but the last one.
   $out =~ s/((\w+\s+)+(?=[^\s,;\[\{\)]))/ ' ' x length $1 /ge;
@@ -495,7 +565,7 @@ sub whited_decl {		# Input is sanitized.
   warn "panic: length mismatch\n\t'$in'\nwhited-out as\n\t'$out'\n"
     if length($in) != length $out;
   # Sanity check
-  warn "panic: multiple types without interwening comma in\n\t$in\nwhited-out as\n\t$out\n"
+  warn "panic: multiple types without intervening comma in\n\t$in\nwhited-out as\n\t$out\n"
     if $out =~ /\w[^\w,]+\w/;
   warn "panic: no types found in\n\t$in\nwhited-out as\n\t$out\n"
     unless $out =~ /\w/;
@@ -531,7 +601,7 @@ sub sanitize {		# We expect that no strings are around
 	     | '((?:[^\\\']|\\.)+)'	# (2) Character constants
 	     | "((?:[^\\\"]|\\.)*)"	# (3) Strings
 	     | ( ^ \s* \# .* 		# (4) Preprocessor
-		 ( \\ $ .* )* )		# and continuation lines
+		 ( \\ $ \n .* )* )	# and continuation lines
 	    } {
 	      # We want to preserve the length, so that one may go back
 	      defined $1 ? ' ' x (1 + length $1) :
@@ -559,6 +629,7 @@ sub top_level {			# We expect argument is sanitized
 
 sub remove_type_decl {		# We suppose that the arg is top-level only.
   my $in = shift;
+  $in =~ s/(\b__extension__)(\s+typedef\b)/(' ' x length $1) . $2/gse;
   $in =~ s/(\btypedef\b.*?;)/' ' x length $1/gse;
   # The following form may appear only in the declaration of the type itself:
   $in =~ 
@@ -893,6 +964,12 @@ Value: a reference to a list of C<typedef>ed names.  Heuristics are used.
 
 Value: a reference to a list of C<extern> variable declarations.
 
+=item C<typedef_structs>
+
+Value: a reference to a hash of parsed struct declarations from typedefs.
+Keys are typedefed names, values are C<undef>s if not a struct, and array
+references (with what?) for values.
+
 =back
 
 =cut
